# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

import sys
from pathlib import Path

from ebookrack.calibre.library import Library


def print_library(lib_path):
    library = Library(lib_path)

    for b in library.books:
        print_book(b)


def print_book(book):
    print(book.title)
    print(" & ".join(book.authors))

    for file in book.files:
        print(f"{file.format:6s}: {file}")

    print()


if __name__ == "__main__":
    try:
        lib_path = Path(sys.argv[1])
    except IndexError:
        print("missing argument: path to library", file=sys.stderr)
        sys.exit(1)

    print_library(lib_path)



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

import xml.etree.ElementTree as ET
from pathlib import Path

import pytest

from ebookrack.calibre.metadata import (
    Metadata,
    parse_date,
    parse_ids,
    parse_opf,
    parse_year,
)

DATA_DIR = Path("tests", "library")
ELVES_OPF = DATA_DIR.joinpath(
    "Andrzej Sapkowski",
    "Blood of Elves (219)",
    "metadata.opf",
)
ORIGIN_OPF = DATA_DIR.joinpath(
    "Charles Darwin",
    "On the Origin of Species By Means of Natural Selection, or, the Preservation of Favoured Races (195)",
    "metadata.opf",
)
ALICE_OPF = DATA_DIR.joinpath(
    "Lewis Carroll",
    "Alice's Adventures in Wonderland (1)",
    "metadata.opf",
)
PLANTS_OPF = DATA_DIR.joinpath(
    "Charles Darwin",
    "Insectivorous Plants (2)",
    "metadata.opf",
)
HOMER_OPF = DATA_DIR.joinpath(
    "Homer",
    "The Odyssey (3)",
    "metadata.opf",
)
opf_files = [ELVES_OPF, ORIGIN_OPF, ALICE_OPF, PLANTS_OPF, HOMER_OPF]

elves_data = Metadata(
    title="Blood of Elves",
    authors=["Andrzej Sapkowski"],
    pubdate="2010",
    publisher="Orion",
    language="eng",
    tags={"Fiction"},
    ids={"MOBI-ASIN": "B0043M66Z4"},
    series="The Witcher",
    series_index="3",
    importdate="2020-11-12",
)
origin_data = Metadata(
    title=(
        "On the Origin of Species By Means of Natural Selection, or, "
        "the Preservation of Favoured Races in the Struggle for Life"
    ),
    authors=["Charles Darwin"],
    pubdate="2009",
    publisher="Project Gutenberg",
    language="eng",
    tags={"Non-fiction", "Biology"},
    ids={"URI": "http://www.gutenberg.org/ebooks/1228"},
    series="",
    series_index="",
    importdate="2019-12-09",
)
alice_data = Metadata(
    title="Alice's Adventures in Wonderland",
    authors=["Lewis Carroll"],
    pubdate="2008",
    publisher="",
    language="eng",
    tags={"Children's stories", "Fantasy fiction"},
    ids={},
    series="",
    series_index="",
    importdate="2023-06-22",
)
plants_data = Metadata(
    title="Insectivorous Plants",
    authors=["Charles Darwin"],
    pubdate="2004",
    publisher="",
    language="eng",
    tags={"Carnivorous plants"},
    ids={},
    series="",
    series_index="",
    importdate="2023-06-22",
)
homer_data = Metadata(
    title="The Odyssey",
    authors=["Homer", "Samuel Butler"],
    pubdate="1999",
    publisher="",
    language="eng",
    tags={"Epic poetry"},
    ids={},
    series="",
    series_index="",
    importdate="2023-06-22",
)
data_sets = [elves_data, origin_data, alice_data, plants_data, homer_data]


@pytest.mark.parametrize("opf_file,data", zip(opf_files, data_sets))
def test_parse_opf(opf_file, data):
    assert parse_opf(opf_file) == data


@pytest.mark.parametrize(
    "timestamp,date",
    [
        ("2020-11-12T11:00:58+00:00", "2020-11-12"),
        ("", ""),
    ],
)
def test_parse_date(timestamp, date):
    assert parse_date(timestamp) == date


@pytest.mark.parametrize(
    "timestamp,year",
    [
        ("2020-11-12T11:00:58+00:00", "2020"),
        ("", ""),
    ],
)
def test_parse_year(timestamp, year):
    assert parse_year(timestamp) == year


ids_xml = """\
<metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:opf="http://www.idpf.org/2007/opf">
  <dc:identifier opf:scheme="calibre" id="calibre_id">219</dc:identifier>
  <dc:identifier opf:scheme="uuid" id="uuid_id">29a8f81c-ca4f-43d5-9638-535936ed62d5</dc:identifier>
  <dc:identifier opf:scheme="MOBI-ASIN">B0043M66Z4</dc:identifier>
  <dc:identifier opf:scheme="URI">http://www.gutenberg.org/ebooks/1228</dc:identifier>
  <dc:identifier opf:scheme="ISBN">9781680508451</dc:identifier>
</metadata>
"""


def test_parse_ids():
    meta = ET.fromstring(ids_xml)
    ids = meta.findall("{http://purl.org/dc/elements/1.1/}identifier")
    assert parse_ids(ids) == {
        "MOBI-ASIN": "B0043M66Z4",
        "ISBN": "9781680508451",
        "URI": "http://www.gutenberg.org/ebooks/1228",
    }



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

from pathlib import Path

import pytest

from ebookrack.calibre.library import Book, Library

DATA_DIR = Path("tests", "library")
BOOK_DIR = DATA_DIR / "Charles Darwin" / "Insectivorous Plants (2)"


def test_library_has_list_of_books():
    library = Library(DATA_DIR)
    assert len(library.books) == 5


@pytest.mark.parametrize("field", ["title", "authors", "path", "files", "cover"])
def test_book_has_field(field):
    book = Book(BOOK_DIR)
    assert hasattr(book, field)


def test_book_path_is_relative_to_library_root():
    book = Book(BOOK_DIR)
    assert DATA_DIR / book.path == BOOK_DIR


def test_book_files_have_format():
    book = Book(BOOK_DIR)
    assert {f.format for f in book.files} == {"EPUB", "PDF"}



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

import enum

import attr

__all__ = ["Catalogue"]


Status = enum.Enum("Status", "ACTIVE DELETED")


@attr.s
class Card:
    _book_id = attr.ib()
    _data = attr.ib()
    _status = attr.ib(default=Status.ACTIVE)

    @property
    def book_data(self):
        return {"id": self._book_id, "status": str(self._status.name), **self._data}

    @property
    def is_deleted(self):
        return self._status is Status.DELETED

    def matches(self, params):
        for field, search_term in params.items():
            if field not in self._data:
                return False

            if isinstance(search_term, set):
                if not search_term.issubset(self._data[field]):
                    return False
            else:
                if self._data[field] != search_term:
                    return False

        return True

    def edit(self, new_data):
        self._data.update(new_data)

    def erase(self):
        self._data = {}
        self._status = Status.DELETED


@attr.s(frozen=True)
class Catalogue:
    _cards = attr.ib(factory=list)

    @property
    def card_index(self):
        return {
            card.book_data["id"]: card._data
            for card in self._cards
            if not card.is_deleted
        }

    def new_card(self, book_data):
        card = Card(book_id=len(self._cards), data=book_data)
        self._cards.append(card)
        return card.book_data["id"]

    def search(self, **query):
        return [card._data for card in self._cards if card.matches(query)]

    def retrieve_card(self, book_id):
        return self._cards[book_id].book_data

    def edit_card(self, book_id, **new_data):
        card = self._cards[book_id]
        card.edit(new_data)

    def destroy_card(self, book_id):
        card = self._cards[book_id]
        card.erase()



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

import shutil
from pathlib import Path

import pytest

from ebookrack.shelf import Book, Shelf, slugify


def test_slugify():
    assert slugify("Hello, World…/ ") == "hello-world"


@pytest.fixture
def input_file(tmpdir):
    upload_dir = Path(tmpdir) / "uploads"
    upload_dir.mkdir()
    book_path = upload_dir / "TheInferno.epub"
    book_path.write_text("# The Inferno\n")
    return book_path


@pytest.fixture
def shelf(tmpdir):
    return Shelf(Path(tmpdir))


def test_shelve_renames_file_and_retains_original(shelf, input_file):
    shelf.shelve(1, dict(title="The Inferno"), input_file)
    assert shelf.locate(1) == [Book(shelf._root_dir / "1" / "the-inferno.epub")]
    assert (shelf._root_dir / "1" / "the-inferno.epub").is_file()
    assert input_file.is_file()


def test_reshelve_moves_file_and_updates_location_table(shelf, input_file):
    shelf.shelve(1, dict(title="The Inferno"), input_file)
    shelf.reshelve(1, dict(title="Inferno"))

    assert shelf.locate(1) == [Book(shelf._root_dir / "1" / "inferno.epub")]
    assert (shelf._root_dir / "1" / "inferno.epub").is_file()
    assert (shelf._root_dir / "1" / "the-inferno.epub").is_file() is False


def test_discarding_book_deletes_files_and_record(shelf, input_file):
    input_file2 = shutil.copy2(input_file, input_file.with_suffix(".pdf"))
    shelf.shelve(1, dict(title="The Inferno"), input_file)
    shelf.shelve(1, dict(title="The Inferno"), input_file2)
    stored_books = shelf.locate(1)
    assert len(stored_books) == 2
    assert all(book.path.is_file() for book in stored_books)

    shelf.discard(1)

    assert not any(book.path.is_file() for book in stored_books)
    assert shelf.locate(1) == []



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

"""Library component

The Library is the interface by which users (librarians and borrowers)
interact with the contents of the library (books and their metadata).

The following use cases can be performed by a web app, API, CLI, etc.
through this interface:

   #. Acquire new book (add metadata & upload file/s)
   #. Add / update book metadata
   #. Move book to different storage location
   #. Add (upload) multiple formats to a book
   #. Discard a book (metadata & file/s)
   #. Search for a book / certain books
   #. List all books in the library
   #. Download / locate a book
"""

import logging

import attr

from .catalogue import Catalogue
from .shelf import Shelf

__all__ = ["Library"]


@attr.s
class Library:
    _catalogue: Catalogue = attr.ib()
    _shelf: Shelf = attr.ib()

    def acquire(self, book_data, book_file):
        book_id = self._catalogue.new_card(book_data)
        book_data = self._catalogue.retrieve_card(book_id)
        self._shelf.shelve(book_id, book_data, book_file)
        return book_id

    def edit_metadata(self, book_id, **new_data):
        self._catalogue.edit_card(book_id, **new_data)

        if "title" in new_data:
            self.reshelve(book_id)

    def reshelve(self, book_id):
        book_data = self._catalogue.retrieve_card(book_id)
        self._shelf.reshelve(book_id, book_data)

    def acquire_additional_format(self, book_id, book_file):
        book_data = self._catalogue.retrieve_card(book_id)
        self._shelf.shelve(book_id, book_data, book_file)

    def replace_shelf(self, new_shelf):
        self._shelf.transfer_all(self._catalogue.card_index, new_shelf)
        self._shelf = new_shelf

    def discard(self, book_id):
        self._shelf.discard(book_id)
        self._catalogue.destroy_card(book_id)

    def search(self, **query):
        matches = self._catalogue.search(**query)
        logging.info(f"Found {format_count(len(matches), 'book')} matching {query}")
        return matches

    def inventory(self):
        return list(self._catalogue.card_index.values())

    def locate(self, book_id):
        books = self._shelf.locate(book_id)
        return [book.path for book in books]


def format_count(number, noun):
    suffix = "s" if number != 1 else ""
    return f"{number} {noun}{suffix}"



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

from pathlib import Path
from zlib import adler32

import pytest

from ebookrack.catalogue import Catalogue
from ebookrack.library import Library
from ebookrack.shelf import Shelf


def checksum(path):
    return adler32(path.read_bytes())


@pytest.fixture()
def book_data():
    data_inferno = dict(author="Dante", title="The Inferno")
    data_canterb = dict(author="Chaucer", title="The Canterbury tales")
    return [data_inferno, data_canterb]


@pytest.fixture()
def book_files(tmpdir):
    upload_dir = Path(tmpdir) / "uploads"
    upload_dir.mkdir()

    file_inferno = upload_dir / "TheInferno.epub"
    file_inferno.write_text("# The Inferno\n")

    file_canterb = upload_dir / "canterbury+tales.pdf"
    file_canterb.write_text("= Chaucer's Canterbury Tales\n")

    return [file_inferno, file_canterb]


@pytest.fixture()
def library(tmpdir):
    return Library(Catalogue(), Shelf(Path(tmpdir)))
    # TODO: Change initialisation of Library


def test_acquired_books_are_listed_in_inventory(library, book_data, book_files):
    for data, file_ in zip(book_data, book_files):
        library.acquire(data, file_)

    assert library.inventory() == book_data


def test_acquired_books_can_be_located_on_shelf(library, book_data, book_files):
    book_ids = []

    for data, file_ in zip(book_data, book_files):
        book_id = library.acquire(data, file_)
        book_ids.append(book_id)

    for book_id, uploaded_book in zip(book_ids, book_files):
        stored_book = library.locate(book_id)[0]
        assert stored_book.is_file()
        assert checksum(stored_book) == checksum(uploaded_book)


def test_storing_book_in_multiple_formats(library, book_data, book_files):
    inferno_data = book_data[0]
    inferno_epub, tmp_pdf = book_files
    inferno_pdf = inferno_epub.with_suffix(".pdf")
    tmp_pdf.rename(inferno_pdf)

    book_id = library.acquire(inferno_data, inferno_epub)
    library.acquire_additional_format(book_id, inferno_pdf)

    stored_books = library.locate(book_id)
    assert stored_books[1].suffix == ".pdf"
    assert stored_books[1].is_file()
    assert checksum(stored_books[1]) == checksum(inferno_pdf)


def test_transferring_books_to_new_shelf(tmpdir, book_data, book_files):
    root0 = Path(tmpdir) / "initial"
    root1 = Path(tmpdir) / "updated"
    root0.mkdir()
    library = Library(Catalogue(), Shelf(root0))
    book_ids = []

    for data, file_ in zip(book_data, book_files):
        book_id = library.acquire(data, file_)
        book_ids.append(book_id)

    for book_id in book_ids:
        for book_file in library.locate(book_id):
            assert book_file.relative_to(root0)

    root1.mkdir()
    library.replace_shelf(Shelf(root1))

    for book_id in book_ids:
        for book_file in library.locate(book_id):
            assert book_file.relative_to(root1)

    initial_files = [fp for fp in root0.glob("**") if fp.is_file()]
    assert not initial_files


def test_modifying_author_updates_catalogue_card(library, book_data, book_files):
    inferno_data, uploaded_book = book_data[0], book_files[0]
    book_id = library.acquire(inferno_data, uploaded_book)
    initial_path = library.locate(book_id)[0]

    library.edit_metadata(book_id, author="Dante Alighieri", publication_date=1320)

    data = library.search(title="The Inferno")[0]
    assert data["author"] == "Dante Alighieri"
    assert data["publication_date"] == 1320
    assert library.locate(book_id)[0] == initial_path


def test_modifying_title_triggers_reshelving(library, book_data, book_files):
    inferno_data, uploaded_book = book_data[0], book_files[0]
    book_id = library.acquire(inferno_data, uploaded_book)
    initial_path = library.locate(book_id)[0]

    library.edit_metadata(book_id, title="Inferno")

    data = library.search(author="Dante")[0]
    assert data["title"] == "Inferno"

    new_path = library.locate(book_id)[0]
    assert new_path != initial_path
    assert checksum(new_path) == checksum(uploaded_book)


def test_discard_book_destroys_record_and_files(library, book_data, book_files):
    for data, file_ in zip(book_data, book_files):
        book_id = library.acquire(data, file_)

    stored_book = library.locate(book_id)[0]
    assert stored_book.is_file()
    assert len(library.search(title=data["title"])) == 1

    library.discard(book_id)

    assert library.locate(book_id) == []
    assert not stored_book.is_file()
    assert len(library.inventory()) == 1
    assert library.search(title=data["title"]) == []



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

"""Shelf for storing and retrieving books

The Shelf is used to place books in the storage system
and maintains an index of the storage locations.

Books can be shelved, in multiple formats;
moved to a new location on the shelf by providing updated metadata;
moved to a different shelf;
or discarded.

shelf
    index and interface to the storage system
book_id
    the identifier for a book, used throughout the library system
book_data
    the book metadata retrieved from the library catalogue
book
    the e-book itself, in a given format, e.g. epub
"""

import re
import shutil
from collections import defaultdict

import attr

__all__ = ["Shelf"]


@attr.s(frozen=True)
class Shelf:
    _root_dir = attr.ib()
    _books = attr.ib(factory=lambda: defaultdict(list))

    def shelve(self, book_id, book_data, book_file):
        book = Book(book_file)
        location = self.prepare_location(book_id, book_data, book)
        shelved_book = book.copy_to(location)
        self._books[book_id].append(shelved_book)

    def reshelve(self, book_id, book_data):
        books = self._books.pop(book_id)

        for book in books:
            location = self.prepare_location(book_id, book_data, book)
            shelved_book = book.move_to(location)
            self._books[book_id].append(shelved_book)

    def prepare_location(self, book_id, book_data, book):
        basename = slugify(book_data["title"]) + book.format
        return self._root_dir / str(book_id) / basename

    def transfer_all(self, card_index, new_shelf):
        for book_id in list(self._books):
            book_data = card_index[book_id]
            books = self._books.pop(book_id)

            for book in books:
                new_shelf.shelve(book_id, book_data, book.path)
                book.delete()

    def locate(self, book_id):
        return self._books[book_id]

    def discard(self, book_id):
        for book in self._books.pop(book_id):
            book.delete()


@attr.s(frozen=True)
class Book:
    _path = attr.ib()

    @property
    def path(self):
        return self._path

    @property
    def format(self):
        return self._path.suffix

    def copy_to(self, dest):
        dest.parent.mkdir(exist_ok=True)
        shutil.copy2(self._path, dest)
        return Book(dest)

    def move_to(self, dest):
        self._path.rename(dest)
        return Book(dest)

    def delete(self):
        self._path.unlink()


def slugify(string):
    """Convert the string to lowercase and dash-separated.

    Loosely based on: https://github.com/zacharyvoase/slugify
    """
    return re.sub(r"[-\s]+", "-", re.sub(r"[^\w\s-]", "", string).strip().lower())



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2
"""
list the books in an e-book library
"""

from pathlib import Path

from .metadata import parse_opf


class Library:
    def __init__(self, root_dir):
        self.books = [Book(d) for d in root_dir.glob("*/*")]


class Book:
    def __init__(self, book_dir):
        self.metadata = parse_opf(book_dir / "metadata.opf")

        dir_depth = 2  # how many path segments to get to library root
        self.path = "/".join(book_dir.parts[-dir_depth:])

        self.files = find_ebooks(book_dir)
        self.cover = book_dir / "cover.jpg"

    def __getattr__(self, name):
        if hasattr(self.metadata, name):
            return getattr(self.metadata, name)

        return super().__getattr__(self, name)


class BookFile(Path):
    @property
    def format(self):
        return self.suffix.lstrip(".").upper()


def find_ebooks(book_dir):
    return [BookFile(p) for p in book_dir.glob("*.*") if is_ebook(p)]


def is_ebook(p):
    return p.is_file() and p.name not in {"metadata.opf", "cover.jpg"}



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2
"""
parse book metadata from an OPF file
"""

import functools
import xml.etree.ElementTree as ET
from dataclasses import dataclass

nss = {
    "dc": "http://purl.org/dc/elements/1.1/",
    "opf": "http://www.idpf.org/2007/opf",
}


@dataclass(frozen=True)
class Metadata:
    title: str
    authors: list[str]
    pubdate: str
    publisher: str
    language: str
    tags: set[str]
    ids: dict[str, str]
    series: str
    series_index: str
    importdate: str


def parse_date(timestamp):
    return timestamp.split("T")[0]


def parse_year(timestamp):
    return parse_date(timestamp)[:4]


def parse_ids(identifiers):
    """Extract relevant IDs as a scheme:ID mapping.

    Expected input: list of <dc:identifiers> elements
    """
    return {
        scheme: el.text
        for el in identifiers
        if (scheme := el.get("{http://www.idpf.org/2007/opf}scheme"))
        not in {"uuid", "calibre"}
    }


def parse_metadata(metadata):
    find_all = functools.partial(metadata.findall, namespaces=nss)
    find_text = functools.partial(metadata.findtext, namespaces=nss)

    title = find_text("dc:title")
    authors = [x.text for x in find_all("dc:creator")]
    pubdate = parse_year(find_text("dc:date"))
    publisher = find_text("dc:publisher")
    language = find_text("dc:language")
    tags = {x.text for x in find_all("dc:subject")}
    ids = parse_ids(find_all("dc:identifier"))

    extra = {x.get("name"): x.get("content") for x in find_all("opf:meta")}
    series = extra.get("calibre:series", "")
    series_index = extra.get("calibre:series_index", "")
    importdate = parse_date(extra.get("calibre:timestamp", ""))

    return Metadata(
        title or "",
        authors,
        pubdate or "",
        publisher or "",
        language or "",
        tags,
        ids,
        series,
        series_index,
        importdate,
    )


def parse_opf(opf_file):
    tree = ET.parse(opf_file)
    root = tree.getroot()
    meta_elem = root.find("opf:metadata", namespaces=nss)
    return parse_metadata(meta_elem)



# SPDX-FileCopyrightText: 2024 JA Viljoen <ebookrack@javiljoen.net>
# SPDX-License-Identifier: EUPL-1.2

import pytest

from ebookrack.catalogue import Catalogue

inferno = dict(
    ISBN="1234",
    author="Dante",
    title="The Inferno",
    tags={"fiction", "renaissance"},
    language="Italian",
)
life_story = dict(
    ISBN="1235",
    author="Dante",
    translator="Chaucer",
    title="My life story",
    tags={"non-fiction"},
    language="English",
)
canterbury = dict(
    ISBN="1236",
    author="Chaucer",
    title="The Canterbury tales",
    tags={"fiction", "renaissance"},
    language="English",
)


@pytest.fixture
def catalogue():
    catalogue = Catalogue()
    catalogue.new_card(inferno)
    catalogue.new_card(life_story)
    catalogue.new_card(canterbury)
    return catalogue


def test_search_by_author(catalogue):
    assert catalogue.search(author="Dante") == [inferno, life_story]


def test_search_by_single_tag(catalogue):
    assert catalogue.search(tags={"renaissance"}) == [inferno, canterbury]


def test_search_by_multiple_tags(catalogue):
    assert catalogue.search(tags={"fiction", "renaissance"}) == [inferno, canterbury]


def test_search_by_author_and_tag(catalogue):
    assert catalogue.search(author="Dante", tags={"fiction"}) == [inferno]


def test_search_empty_query(catalogue):
    assert catalogue.search() == [inferno, life_story, canterbury]


def test_search_by_unknown_params(catalogue):
    assert catalogue.search(translator="Chaucer") == [life_story]
    assert catalogue.search(reviewer="booksRsux69") == []


def test_search_by_known_and_unknown_param(catalogue):
    assert catalogue.search(author="Dante", translator="Chaucer") == [life_story]
    assert catalogue.search(author="Dante", banned=False) == []
    assert catalogue.search(author="Dante", banned=True) == []


def test_edit_card_updates_card_data():
    catalogue = Catalogue()
    book_id = catalogue.new_card(inferno)

    catalogue.edit_card(book_id, author="Dante Alighieri")

    assert catalogue.retrieve_card(book_id) == {
        **inferno,
        "id": book_id,
        "status": "ACTIVE",
        "author": "Dante Alighieri",
    }


def test_destroy_card_replaces_with_placeholder(catalogue):
    book_id = list(catalogue.card_index).pop()
    catalogue.destroy_card(book_id)

    assert catalogue.retrieve_card(book_id) == {"id": book_id, "status": "DELETED"}


def test_destroyed_cards_are_excluded_from_card_index(catalogue):
    initial_index = catalogue.card_index
    assert list(initial_index) == [0, 1, 2]
    catalogue.destroy_card(1)

    assert catalogue.card_index == {0: initial_index[0], 2: initial_index[2]}
